--[[
	Allows you to keep track of people you have had whisper conversations with.
]]

-- Variables

local TellTrack_ButtonCount = 1
local TellTrack_TooltipSetId = 0
local TellTrack_Array = { }
local TellTrack_ID = 0
local TellTrack_LastTell = nil
local TellTrack_ArrayMaxSize = 20
local TellTrack_ArrayOffset = 1

local TellTrack_WhisperMessages = {}
local TellTrack_CurrentConversation = {name="",showAll=true}
local TellTrack_OfflinePlayers = {}

local TellTrack_Khaos_Registered = nil
local TellTrack_Satellite_Registered = nil

local SavedSendChatMessage = nil
local SavedFCF_SelectDockFrame = nil
local SavedChatFrame_OpenChat = nil
local SavedFCF_Close = nil
local SavedFCF_OpenNewWindow = nil

SELECTED_CHAT_FRAME = DEFAULT_CHAT_FRAME
ERR_CHAT_PLAYER_NOT_FOUND_S_REGEX = gsub(ERR_CHAT_PLAYER_NOT_FOUND_S, "%%s", "(.*)")
ERR_FRIEND_OFFLINE_S_REGEX = gsub(ERR_FRIEND_OFFLINE_S, "%%s", "(.*)")
ERR_FRIEND_ONLINE_SS_REGEX = gsub(ERR_FRIEND_ONLINE_SS, "%%s", "(.*)")

-- executed on load, calls general set-up functions
function TellTrack:OnLoadCustom( frame )
	self._meta.short = "tt"
	
	if (Khaos) then
		if (not TellTrack_Khaos_Registered) then
			self:Register_Khaos()
		end
	end
	if (Satellite) then
		if (not TellTrack_Satellite_Registered) then
			self:Register_Satellite()
		end
	else
		SLASH_TELLTRACK3 = "/ttrack"
		SlashCmdList["TELLTRACKRETELL"] = TellTrack_WhisperToPreviousTarget
		SLASH_TELLTRACKRETELL1 = "/re"
		SLASH_TELLTRACKRETELL2 = "/retell"
	end
        
	frame:RegisterEvent("PLAYER_ENTERING_WORLD")
	frame:RegisterEvent("CHAT_MSG_WHISPER")
	frame:RegisterEvent("CHAT_MSG_WHISPER_INFORM")
	frame:RegisterEvent("CHAT_MSG_SYSTEM")

	self:UpdateTellTrackButtonsText()
	
	-- Hook the chat event handler 'before'
	if (SendChatMessage ~= SavedSendChatMessage) then
		SavedSendChatMessage = SendChatMessage
		SendChatMessage = TellTrack_SendChatMessage
	end
	-- Hook dock frame selection 'after'
	if (FCF_SelectDockFrame ~= TellTrack_FCF_SelectDockFrame) then
		SavedFCF_SelectDockFrame = FCF_SelectDockFrame
		FCF_SelectDockFrame = TellTrack_FCF_SelectDockFrame
	end
	-- Hook open chat to open as whisper
	if (ChatFrame_OpenChat ~= TellTrack_ChatFrame_OpenChat) then
		SavedChatFrame_OpenChat = ChatFrame_OpenChat
		ChatFrame_OpenChat = TellTrack_ChatFrame_OpenChat
	end
	-- Hook open chat to open as whisper
	if (FCF_Close ~= TellTrack_FCF_Close) then
		SavedFCF_Close = FCF_Close
		FCF_Close = TellTrack_FCF_Close
	end
	-- Hook open chat to open as whisper
	if (FCF_OpenNewWindow ~= TellTrack_FCF_OpenNewWindow) then
		SavedFCF_OpenNewWindow = FCF_OpenNewWindow
		FCF_OpenNewWindow = TellTrack_FCF_OpenNewWindow
	end
	
	TellTrack_SlashSubCommandList = {
		enabled = { name="enabled", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_Enabled) end, help=TELLTRACK_CHAT_COMMAND_ENABLE_INFO},
		invert = { name="invert", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_Inverted) end, help= TELLTRACK_CHAT_COMMAND_INVERT_INFO},
		dontsavelist = { name="dontsavelist", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_DontSaveList) end, help= TELLTRACK_CHAT_COMMAND_DONTSAVELIST_INFO},
		autocreate = { name="autocreate", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_AutoCreate) end, help= TELLTRACK_CHAT_COMMAND_AUTOCREATE_INFO},
		hidewhispers = { name="hidewhispers", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_HideOtherWhispers) end, help= TELLTRACK_CHAT_COMMAND_HIDEOTHERWHISPERS_INFO},
		timestamps = { name="timestamps", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_TimeStamps) end, help=TELLTRACK_CHAT_COMMAND_TIMESTAMPS_INFO},
		whisperfirst = { name="whisperfirst", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_WhisperFirst) end, help=TELLTRACK_CHAT_COMMAND_WHISPERFIRST_INFO},
		clearall = { name="clearall", func=TellTrack_Menu_DeleteAll, help=TELLTRACK_CHAT_COMMAND_CLEARALL_INFO},
		border = { name="border", func=function(msg) self:ChatCommandToggler(msg, TellTrack_Toggle_HideBorder) end, help=TELLTRACK_CHAT_COMMAND_HIDEBORDER_INFO},
	}
	
end

function TellTrack:OnVariablesLoaded( frame )
	local db,options = self._db,self._options

	self.L = self._acelib.L
	self.LR = self._acelib.LR
	
--	getglobal(frame:GetName() .. "HeaderIcon"):SetTexture( self._meta.icon )
--	getglobal(frame:GetName() .. "HeaderTitle"):SetText( self._meta.title )
--	getglobal(frame:GetName() .. "HeaderVersion"):SetText( self._options.Color .. "v" .. self._meta.version )  -- .. " "..self._meta.locale.."|r" )

	options.Realm = GetRealmName()
	options.Player = UnitName("Player")
	if not db.Chat[options.Realm] then
		db.Chat[options.Realm] = { }
	end
	if not db.Chat[options.Realm][options.Player] then
		db.Chat[options.Realm][options.Player] = { }
	end
	self._chat = db.Chat[options.Realm][options.Player]
	
	if not TellTrack_Khaos_Registered then
		TellTrack_Toggle_Enabled(options.Enabled)
	end
	if TellTrack._options.DontSaveList == 1 then
		db.Log = { }
	end
	self:LoadNames()
	TellTrack_Toggle_UpdateBorderVisibility()
	--Initialize Menu
	UIDropDownMenu_Initialize(TellTrackDropDown, TellTrackTextButton_InitializeMenuArray, "MENU")
end

function TellTrack:OnPlayerLogin( frame )
	local _inst = self._acelib
	if _inst then
		if _inst.menu then
			_inst.menu.args.help = nil
		end
	end
end

function TellTrack:OnCommandCustom( msg, cmd, args)
	if not cmd then
		TellTrack_SlashSubCommandList["enabled"].func(args); return true
	elseif TellTrack_SlashSubCommandList[cmd] then
		TellTrack_SlashSubCommandList[cmd].func(args); return true
	elseif cmd == "help" then
		self:ChatCommand_Help()
	end
end

function TellTrack:Print_Help()
	self:Print(TELLTRACK_CHAT_QUESTION_MARK_INFO)
	self:Print(TELLTRACK_CHAT_QUESTION_MARK_INFO1)
	self:Print(TELLTRACK_CHAT_QUESTION_MARK_INFO2)
	self:Print(TELLTRACK_CHAT_QUESTION_MARK_INFO3)
end

function TellTrack:ChatCommand_Help()
	self:Print(TELLTRACK_CHAT_USE_INFO)
	for subcom, info in pairs(TellTrack_SlashSubCommandList) do
		self:Print("/"..self._meta.short.." "..subcom.." - ".. info.help)
	end
end

function TellTrack:ChatCommandToggler(msg, func)
	msg = msg and strlower(strtrim(msg)) or ""
	
	-- Toggle appropriately
	if ( (string.find(msg, 'on')) or ((string.find(msg, '1')) and (not string.find(msg, '-1')) ) ) then
		func(1)
	elseif ( (string.find(msg, 'off')) or (string.find(msg, '0')) ) then
		func(0)
	else
		func(-1)
	end
end

function TellTrack_ShowAllWhispers()
	TellTrack:ShowConversationWith(TellTrack_CurrentConversation.name, true)
end

-- Toggles the enabled/disabled state of an option and returns the new state
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack:Generic_Toggle(toggle, variableName, enableMessage, disableMessage, KhaosVarName)
	local oldvalue = self._options[variableName]
	local newvalue = toggle
	if ( toggle ~= 1 ) and ( toggle ~= 0 ) then
		if oldvalue == 1 then
			newvalue = 0
		elseif oldvalue == 0 then
			newvalue = 1
		else
			newvalue = 0
		end
	end
	self._options[variableName] = newvalue
	if ( ( newvalue ~= oldvalue ) and ( not TellTrack_Khaos_Registered ) ) then
		if variableName~="Enabled" then
			if ( newvalue == 1 ) then
				self:Print(getglobal(enableMessage))
			else
				self:Print(getglobal(disableMessage))
			end
		end
	end
	if (TellTrack_Khaos_Registered) then
		if (Khaos.getSetKey("TellTrack", KhaosVarName)) then
			Khaos.setSetKeyParameter("TellTrack", KhaosVarName, "checked", (newvalue == 1))
		end
	end
	return newvalue
end

-- Toggles the enabled/disabled state of the TellTrack
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack_Toggle_Enabled(toggle)
	local newvalue = TellTrack:Generic_Toggle(toggle, "Enabled", "TELLTRACK_CHAT_ENABLED", "TELLTRACK_CHAT_DISABLED", "TellTrackEnable")
	if  TellTrack._options.Enabled == 0 then
		TellTrackFrame:Hide()
		if TellTrackTooltip then
			TellTrackTooltip:Hide()
		end
	else
		TellTrackFrame:Show()
	end
end

-- Toggles the inverted/normalized state of the TellTrack list
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack_Toggle_Inverted(toggle)
	local oldvalue = TellTrack._options.InvertedList
	local newvalue = TellTrack:Generic_Toggle(toggle, "InvertedList", "TELLTRACK_CHAT_INVERTED", "TELLTRACK_CHAT_NORMALIZED", "TellTrackInvert")
	--only change (reset offset) telltrack if InvertedList was changed
	if ( oldvalue ~= newvalue ) then
		if ( newvalue == 1 ) then
			TellTrack:ChangeArrayOffset(TellTrack_ArrayMaxSize)
		else
			TellTrack:ChangeArrayOffset(1)
		end
		TellTrack:UpdateTellTrackButtonsText()
	end
end

-- Toggles the saved state of the TellTrack list accross sessions
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack_Toggle_DontSaveList(toggle)
	local newvalue = TellTrack:Generic_Toggle(toggle, "DontSaveList", "TELLTRACK_CHAT_DONTSAVELIST", "TELLTRACK_CHAT_SAVELIST", "TellTrackDontSaveList")
end

-- Toggles the auto creation of TellTrack whisper frame
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack_Toggle_AutoCreate(toggle)
	local newvalue = TellTrack:Generic_Toggle(toggle, "AutoCreate", "TELLTRACK_CHAT_AUTOCREATE", "TELLTRACK_CHAT_MANUALWHISPERFRAME", "TellTrackAutoCreate")
end

-- Toggles Hiding Whispers outside the whisper frame
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack_Toggle_HideOtherWhispers(toggle)
	local newvalue = TellTrack:Generic_Toggle(toggle, "HideOtherWhispers", "TELLTRACK_CHAT_HIDEOTHERWHISPERS", "TELLTRACK_CHAT_SHOWOTHERWHISPERS", "TellTrackHideOtherWhispers")
	if ( newvalue == 1 ) then
		TellTrack:HideWhispersInOtherFrames()
	end
end

-- Toggles Whisper chat stamping
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack_Toggle_TimeStamps(toggle)
	local newvalue = TellTrack:Generic_Toggle(toggle, "TimeStamps", "TELLTRACK_CHAT_TIMESTAMPS", "TELLTRACK_CHAT_NOTIMESTAMPS", "TellTrackTimeStamps")
	TellTrack:RedrawWhisperFrame()
end

-- Toggles the whisper first option
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack_Toggle_WhisperFirst(toggle)
	local newvalue = TellTrack:Generic_Toggle(toggle, "WhisperFirst", "TELLTRACK_CHAT_WHISPERFIRST", "TELLTRACK_CHAT_LOGFIRST", "TellTrackWhisperFirst")
end

-- Toggles the border hiding option
--  if toggle is 1, it's enabled
--  if toggle is 0, it's disabled
--   otherwise, it's toggled
function TellTrack_Toggle_HideBorder(toggle)
	local newvalue = TellTrack:Generic_Toggle(toggle, "HideBorder", "TELLTRACK_CHAT_HIDEBORDER", "TELLTRACK_CHAT_SHOWBORDER", "TellTrackHideBorder")
	TellTrack_Toggle_UpdateBorderVisibility()
end

function TellTrack_Toggle_UpdateBorderVisibility()
	if (TellTrack._options.HideBorder == 1) then
		TellTrackFrameBackground:Hide()
		TellTrackFrameArrowUpButton:Hide()
		TellTrackFrameArrowDownButton:Hide()
		TellTrackFrame:SetHitRectInsets(10, 10, 10+TellTrackFrameArrowUpButton:GetHeight(), 10+TellTrackFrameArrowDownButton:GetHeight())
	else
		TellTrackFrameBackground:Show()
		TellTrackFrameArrowUpButton:Show()
		TellTrackFrameArrowDownButton:Show()
		TellTrackFrame:SetHitRectInsets(0, 0, 0, 0)
	end
end

function TellTrack_WhisperToPreviousTarget(msg)
	if not TellTrack_LastTell then
		return
	end
	SendChatMessage(msg, "WHISPER", this.language, TellTrack_LastTell)
end

function TellTrack_Menu_Whisper()
	TellTrack:InitiateWhisperByID(TellTrack_ID)
end

function TellTrack_Menu_Show_Conversation()
	TellTrack:ShowConversationWith(TellTrack:GetName(TellTrack_ID), false)
end

function TellTrack_Menu_Who()
	local name = TellTrack:GetName(TellTrack_ID)
	SendWho("n-\""..name.."\"")
end

function TellTrack_Menu_GroupeInvite()
	local name = TellTrack:GetName(TellTrack_ID)
	InviteUnit(name)
end

function TellTrack_Menu_AddToFriend()
	local name = TellTrack:GetName(TellTrack_ID)
	AddFriend(name)
end

function TellTrack_Menu_Delete()
	TellTrack:EraseByID(TellTrack_ID)
end

-- Function by AnduinLothar used in right click menu
function TellTrack_Menu_DeleteAll()
	TellTrack_Array = {}
	TellTrack:UpdateTellTrackButtonsText()
	-- saves the names to the list
	TellTrack:SaveNames()
end


-- Handles events
function TellTrack:OnEventCustom(frame, event, ...)
	if ( event == "PLAYER_ENTERING_WORLD" ) then
		self:ModifyButtonCount(TellTrackFrame:GetHeight())
	elseif ( event == "CHAT_MSG_WHISPER" ) or ( event == "CHAT_MSG_WHISPER_INFORM" ) then
		--For some reason self:HookAllAddMessages doesn't work on VARIABLES_LOADED...
		self:HookAllAddMessages()
		
		--Special Hidden Whisper Ignores
        if ( IsAddOnLoaded("Guilded") ) and (string.len(arg1) > 2) and (string.sub(arg1,1,2) == "<G") and ( (string.len(arg1) ~= 4) or (string.sub(arg1,1,4) ~= "<GM>") ) then
			--Hide Guilded messages
			return
		elseif ( IsAddOnLoaded("GuildAds") ) and (string.len(arg1) > 3) and (string.sub(arg1,1,3) == "<GA") then
			--Hide GuildAds messages
			return
		elseif (string.len(arg1) > 3) and (string.sub(arg1,1,3) == "<T>") then
			--Hide Telepathy messages
			return
        elseif (string.len(arg1) > 1) and (string.sub(arg1,1,1) == "/") then
            --Hide AceComm messages
            return
        elseif (string.len(arg1) > 7) and (string.sub(arg1,1,7) == "SKP1019") then
            --Hide SKP messages
            return
		end 
		
		local type = string.sub(event,10)
		if ( event == "CHAT_MSG_WHISPER" ) then
			self:HandleMessageSentOrRecieved(arg2, true, arg1)
		end
		
		local message = gsub(arg1, "%%", "%%%%")
        if (arg2) then
            TellTrack_OfflinePlayers[arg2] = nil
        end
		table.insert(TellTrack_WhisperMessages, {type=type, tag=arg6, name=arg2, msg=message, time=self:GetTime(nil, self._update.timediff)})
	elseif ( event == "CHAT_MSG_SYSTEM" ) then
		local _, _, name = strfind(arg1, ERR_CHAT_PLAYER_NOT_FOUND_S_REGEX)
		if (name) then
			TellTrack_OfflinePlayers[name] = true
			self:UpdateTellTrackButtonsText()
			return
		end
		_, _, name = strfind(arg1, ERR_FRIEND_OFFLINE_S_REGEX)
		if (name) then
			TellTrack_OfflinePlayers[name] = true
			self:UpdateTellTrackButtonsText()
			return
		end
		_, _, name = strfind(arg1, ERR_FRIEND_ONLINE_SS_REGEX)
		if (name) then
			TellTrack_OfflinePlayers[name] = nil
			self:UpdateTellTrackButtonsText()
			return
		end
	end
end



function TellTrack:startsWith(s, prefix)
	local isAtBeginning = false
	if (type(s) == "string") and (type(prefix) == "string") then
		if (s == prefix) then
			isAtBeginning = true
		elseif ( string.len(s) > string.len(prefix) ) then
			if( string.sub(s, 1, string.len(prefix)) == prefix ) then
				isAtBeginning = true
			end
		end
	end
	return isAtBeginning
end


function TellTrack:AddWhisper(chatFrame, data)
	local info = ChatTypeInfo[data.type]
	
	local part1 = getglobal("CHAT_"..data.type.."_GET")..data.msg
	local part2 = (data.tag and _G["CHAT_FLAG_"..data.tag] or "").."|Hplayer:"..data.name.."|h".."["..data.name.."]".."|h"
	if TasteTheNaimbow_Loaded then
		part2 = "|cff"..TasteTheNaimbowExternalColor(name)..part2.."|r"
	end

	chatFrame:AddMessage(format(part1, part2), info.r, info.g, info.b, info.id, nil, data.time)
end


function TellTrack:GetTimeStamp(ts)
	local separator = "" -- ">"
	if ChatTimeStamps_TimeStamp and not ChatTimeStamps_SEPARATOR then
		separator = ""
	end
	if not ts then
		if ChatTimeStamps_LOCALTIME then
			ts = date()
		elseif Clock_GetTimeText then
			ts = Clock_GetTimeText()
		else
			ts = self:GetTime(nil, self._update.timediff)
		end
	end
	if type(ts) == "number" then
		ts = date("%H:%M:%S", ts)
	end	
	if type(ts) ~= "string" then
		ts = ""
	end
	return ts..separator.." "
end


function TellTrack:ShowConversationWith(name, showAll)
	local db,options,chat = self._db,self._options,self._chat
	if not chat.frame or not getglobal(chat.frame) then
		if options.AutoCreate==0 or not self:CreateWhisperChatFrame() then
			return
		end
	end
	
	local WhisperFrame = getglobal(chat.frame)
	TellTrack_CurrentConversation = {name=name,showAll=true}
	WhisperFrame:Clear()
	TellTrack_CurrentConversation.showAll = showAll
	local found = 0
	--Display only the valid whispers.
	for i, data in ipairs(TellTrack_WhisperMessages) do
		if (showAll) or (data.name == name) then
			self:AddWhisper(WhisperFrame, data)
			found = found + 1
		end
	end
	
	if found == 0 then
		TellTrack_CurrentConversation.showAll = true
		local timeStamp = options.TimeStamps
		options.TimeStamps = 0
		WhisperFrame:AddMessage(format(TELLTRACK_NO_CURRENT_CONVERSATION, name))
		options.TimeStamps = timeStamp
	end
	chat.name = name
	
	-- Show the frame and tab
	FCF_SetWindowName(WhisperFrame, chat.name)
    if WhisperFrame.isDocked then
		FCF_SelectDockFrame(WhisperFrame)
    end
end

function TellTrack_WhisperChatFrame_AddMessage(frame, text, r, g, b, id, addToStart, ts)
	if frame:GetName() ~= TellTrack._chat.frame then
		-- incorrect whisper frame
		SavedWhisperChatFrame_AddMessages[frame:GetName()](frame, text, r, g, b, id, addToStart)
		return
	end
	local stamps = ChatTimeStamps_ENABLED
	ChatTimeStamps_ENABLED = 0
	if (strlen(text) < 5) then
		SavedWhisperChatFrame_AddMessages[frame:GetName()](frame, text, r, g, b, id, addToStart)
		ChatTimeStamps_ENABLED = stamps
		return
	end
	local timedText = text
	if (TellTrack._options.TimeStamps == 1) then
		timedText = TellTrack:GetTimeStamp(ts)..text
	end
	if (not TellTrack_CurrentConversation) or (TellTrack_CurrentConversation.showAll) then
		SavedWhisperChatFrame_AddMessages[frame:GetName()](frame, timedText, r, g, b, id, addToStart)
		ChatTimeStamps_ENABLED = stamps
		return
	end
	local name = TellTrack_CurrentConversation.name
    local namecolor = "ffffff"
    if (TasteTheNaimbow_Loaded) then
        namecolor = TasteTheNaimbowExternalColor(name)
    end
    --DEFAULT_CHAT_FRAME:AddMessage("Real "..string.gsub(string.gsub(text,":%d+\124","\124"),"\124","\\124"))
    --DEFAULT_CHAT_FRAME:AddMessage("Fake "..string.gsub(format(CHAT_WHISPER_GET, "|Hplayer:"..name.."|h".."["..name.."]".."|h"),"\124","\\124"))
    --DEFAULT_CHAT_FRAME:AddMessage(tostring((TellTrack:startsWith(string.gsub(text,":%d+\124","\124"), format(CHAT_WHISPER_GET, "|Hplayer:"..name.."|h".."["..name.."]".."|h")))))
    local matchText = string.gsub(text,":%d+\124","\124")
	if  (TellTrack:startsWith(matchText, format(CHAT_WHISPER_GET, "|Hplayer:"..name.."|h".."["..name.."]".."|h"))) or
        (TellTrack:startsWith(matchText, format(CHAT_WHISPER_GET, CHAT_FLAG_DND.."|Hplayer:"..name.."|h".."["..name.."]".."|h"))) or
        (TellTrack:startsWith(matchText, format(CHAT_WHISPER_GET, CHAT_FLAG_AFK.."|Hplayer:"..name.."|h".."["..name.."]".."|h"))) or
        (TellTrack:startsWith(matchText, format(CHAT_WHISPER_GET, CHAT_FLAG_GM.."|Hplayer:"..name.."|h".."["..name.."]".."|h"))) or
		(TellTrack:startsWith(matchText, format(CHAT_WHISPER_INFORM_GET, "|Hplayer:"..name.."|h".."["..name.."]".."|h"))) or
        (TellTrack:startsWith(matchText, format(CHAT_WHISPER_GET, "|cff"..namecolor.."|Hplayer:"..name.."|h".."["..name.."]".."|h|r"))) or
        (TellTrack:startsWith(matchText, format(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_DND.."|Hplayer:"..name.."|h".."["..name.."]".."|h|r"))) or
        (TellTrack:startsWith(matchText, format(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_AFK.."|Hplayer:"..name.."|h".."["..name.."]".."|h|r"))) or
        (TellTrack:startsWith(matchText, format(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_GM.."|Hplayer:"..name.."|h".."["..name.."]".."|h|r"))) or
        (TellTrack:startsWith(matchText, format(CHAT_WHISPER_INFORM_GET, "|cff"..namecolor.."|Hplayer:"..name.."|h".."["..name.."]".."|h|r"))) or
        (TellTrack:startsWith(matchText, format(CHAT_WHISPER_GET, "[".."|Hplayer:"..name.."|h"..name.."|h".."]"))) or
        (TellTrack:startsWith(matchText, format(CHAT_WHISPER_GET, CHAT_FLAG_DND.."[".."|Hplayer:"..name.."|h"..name.."|h".."]"))) or
        (TellTrack:startsWith(matchText, format(CHAT_WHISPER_GET, CHAT_FLAG_AFK.."[".."|Hplayer:"..name.."|h"..name.."|h".."]"))) or
        (TellTrack:startsWith(matchText, format(CHAT_WHISPER_GET, CHAT_FLAG_GM.."[".."|Hplayer:"..name.."|h"..name.."|h".."]"))) or
		(TellTrack:startsWith(matchText, format(CHAT_WHISPER_INFORM_GET, "[".."|Hplayer:"..name.."|h"..name.."|h".."]"))) or
        (TellTrack:startsWith(matchText, format(CHAT_WHISPER_GET, "|cff"..namecolor.."[".."|Hplayer:"..name.."|h"..name.."|h".."]".."|r"))) or
        (TellTrack:startsWith(matchText, format(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_DND.."[".."|Hplayer:"..name.."|h"..name.."|h".."]".."|r"))) or
        (TellTrack:startsWith(matchText, format(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_AFK.."[".."|Hplayer:"..name.."|h"..name.."|h".."]".."|r"))) or
        (TellTrack:startsWith(matchText, format(CHAT_WHISPER_GET, "|cff"..namecolor..CHAT_FLAG_GM.."[".."|Hplayer:"..name.."|h"..name.."|h".."]".."|r"))) or
        (TellTrack:startsWith(matchText, format(CHAT_WHISPER_INFORM_GET, "|cff"..namecolor.."[".."|Hplayer:"..name.."|h"..name.."|h".."]".."|r"))) then
		SavedWhisperChatFrame_AddMessages[frame:GetName()](frame, timedText, r, g, b, id, addToStart)
		ChatTimeStamps_ENABLED = stamps
		return
	end
end


function TellTrack:RedrawWhisperFrame( )	
	if not self._chat.frame then return end
	local WhisperFrame = getglobal(self._chat.frame)
	local scrollOffset = WhisperFrame:GetCurrentScroll()
	WhisperFrame:Clear()
	for i, data in pairs(TellTrack_WhisperMessages) do
		if (TellTrack_CurrentConversation.showAll) or (data.name == TellTrack_CurrentConversation.name) then
			self:AddWhisper(WhisperFrame, data)
		end
	end
	WhisperFrame:SetScrollOffset(scrollOffset)
end

function TellTrack:HookAllAddMessages()
	if SavedWhisperChatFrame_AddMessages then
		return
	end
	SavedWhisperChatFrame_AddMessages = {}
	local frameName = ""
	for i=1, NUM_CHAT_WINDOWS do
		frameName = "ChatFrame"..i
		if (getglobal(frameName).AddMessage ~= TellTrack_WhisperChatFrame_AddMessage) then
			SavedWhisperChatFrame_AddMessages[frameName] = getglobal(frameName).AddMessage
			getglobal(frameName).AddMessage = TellTrack_WhisperChatFrame_AddMessage
		end
	end
end
		
function TellTrack:FindChatFrame( ... )
	for i=1, NUM_CHAT_WINDOWS do
		local sChatFrame = "ChatFrame"..i
		local chatFrame = getglobal(sChatFrame)
		local sName, _, _, _, _, _, bShown = GetChatWindowInfo(i)
		if select("#",...) == 0 then
			if (not bShown and not chatFrame.isDocked) or i == NUM_CHAT_WINDOWS then
				return chatFrame
			end	
		elseif bShown or chatFrame.isDocked then
			for i1 = 1, select("#",...) do
				local name = select(i1,...)
				if name and name == sName then
					return sChatFrame,sName
				end
			end
		end
	end
end 

function TellTrack:CreateWhisperChatFrame()
	local chat,default,name = self._chat,TELLTRACK_WHISPER_FRAME_TITLE
	if chat.frame and getglobal(chat.frame) then return chat.frame end
	
	chat.frame,name = self:FindChatFrame(chat.name, chat.name~=default and default);	-- get if already exists
	if not chat.frame then
		if FCF_GetNumActiveChatFrames() >= NUM_CHAT_WINDOWS then return end				-- dont get if have to take someone elses
		name = chat.name or default
		local selected = SELECTED_DOCK_FRAME
		FCF_OpenNewWindow(name)
		FCF_SelectDockFrame(selected)
		chat.frame = self:FindChatFrame(name)
	end
	if chat.frame then																	-- make sure clear out
		chat.name = name
		local chatFrame = getglobal(chat.frame)
		ChatFrame_RemoveAllMessageGroups(chatFrame)
		ChatFrame_AddMessageGroup(chatFrame, "WHISPER")
		return true
	end
end

function TellTrack:HideWhispersInOtherFrames()
	local chat = self._chat
	if not chat.frame or not getglobal(chat.frame) then	return end
	for i=1, NUM_CHAT_WINDOWS do
		local sChatFrame = "ChatFrame"..i
		local sName, _, _, _, _, _, bShown = GetChatWindowInfo(i)
		if (bShown or chatFrame.isDocked) and sChatFrame ~= chat.frame then
			ChatFrame_RemoveMessageGroup(getglobal(sChatFrame), "WHISPER")
		end
	end
end


-- Does things with the hooked function
function TellTrack_SendChatMessage(text, type, language, target)
	TellTrack:SendChatMessage(text, type, language, target)
end

function TellTrack:SendChatMessage(text, type, language, target)
	-- saves target for 'Retell' regardless of Enabled
	if target then
		if string.find(target, "^[a-zA-Z].*") then 
			target = strupper(strsub(target,1,1))..strlower(strsub(target,2))
		else
			target = strupper(strsub(target,1,2))..strlower(strsub(target,3))
		end
	end

	if ( type == "WHISPER" ) then
		TellTrack_LastTell = target
	end
	if ( self._options.Enabled == 1 ) then
		if ( type == "WHISPER" ) then 
			-- prevent data message transfers from being used
			if ( strsub(text, 1, 1) ~= "<" ) then
				TellTrack:HandleMessageSentOrRecieved(target, false)
			end
		end
	end
	if (SavedSendChatMessage) then
		SavedSendChatMessage(text, type, language, target)
	end
end

function TellTrack_ChatFrame_OpenChat(text, chatFrame)
	local chat = TellTrack._chat
	if (text == "") and (SELECTED_CHAT_FRAME:GetName() == chat.frame) and (TellTrack_CurrentConversation)
			and (TellTrack_CurrentConversation.name) and (TellTrack_CurrentConversation.name ~= "") then
		TellTrack:InitiateWhisperToTarget(TellTrack_CurrentConversation.name)
	else
		if SavedChatFrame_OpenChat then
			SavedChatFrame_OpenChat(text, chatFrame)
		end
	end
end

function TellTrack_FCF_OpenNewWindow( name )
	local isHide = (TellTrack._options.HideOtherWhispers == 1)
	local chatFrame = (isHide) and TellTrack:FindChatFrame() or nil
	if SavedFCF_OpenNewWindow then
		SavedFCF_OpenNewWindow(name)
		if isHide and chatFrame then
			ChatFrame_RemoveMessageGroup(chatFrame, "WHISPER")
		end
	end
end

function TellTrack_FCF_Close( chatFrame, fallBack )
	local frame = fallBack or chatFrame or FCF_GetCurrentChatFrame()
	if frame and frame:GetName() == TellTrack._chat.frame then
		TellTrack._chat.frame = nil
	end
	if SavedFCF_Close then
		SavedFCF_Close(chatFrame, fallBack)
	end
end

function TellTrack_FCF_SelectDockFrame(frame)
	if SavedFCF_SelectDockFrame then
		SavedFCF_SelectDockFrame(frame)
	end
	if (frame:GetName() == TellTrack._chat.frame ) then
		if (TellTrack_CurrentConversation.showAll) then
			for i = 1, TellTrack_ArrayMaxSize do
				if (TellTrack_Array[i]) then
					TellTrack_Array[i].unread = 0
				end
			end
		elseif (TellTrack_CurrentConversation.name ~= "") then
			for i = 1, TellTrack_ArrayMaxSize do
				if (TellTrack_Array[i]) and (TellTrack_Array[i].name == TellTrack_CurrentConversation.name) then
					TellTrack_Array[i].unread = 0
				end
			end
		end
		TellTrack:UpdateTellTrackButtonsText()
	end
	SELECTED_CHAT_FRAME = frame;		--FIXME: is this necessary
end

function TellTrack:ArrowButton_OnClick( frame, button, isUp )
	if not frame:GetID() then return end
	if button == "RightButton" then
		self:ChangeArrayOffset( isUp and 1 or TellTrack_ArrayMaxSize)
	elseif button == "LeftButton" then
		self:ChangeArrayOffset( TellTrack_ArrayOffset + (TellTrack_ButtonCount * (isUp and -1 or 1)) )
	end
end

function TellTrack:ArrowButton_OnSizeChanged( frame )
	local height = frame:GetHeight()
	getglobal(frame:GetName().."RightBG"):SetHeight(height)
	getglobal(frame:GetName().."LeftBG"):SetHeight(height)
	getglobal(frame:GetName().."Arrow"):SetHeight(height)
end


function TellTrack:InitiateWhisperByID(id)
	local name = self:GetName(id)
	self:InitiateWhisperToTarget(name)
end

-- 1/20/05 'Retell' added by AnduinLothar
function TellTrack:InitiateWhisperToPreviousTarget()
	if not TellTrack_LastTell then
		return
	end
	self:InitiateWhisperToTarget(TellTrack_LastTell)
end

function TellTrack:InitiateWhisperToTarget(name)
	if not name then
		return
	end
	local chatFrame = DEFAULT_CHAT_FRAME
	chatFrame.editBox:SetAttribute("chatType", "WHISPER")
	chatFrame.editBox:SetAttribute("tellTarget", name)
	ChatEdit_UpdateHeader(chatFrame.editBox)
	if not chatFrame.editBox:IsVisible() then
		if SavedChatFrame_OpenChat then
			SavedChatFrame_OpenChat("", chatFrame)
		end
	end
end



-- Modified by AnduinLothar to add "Delete All", and "Invert List"

function TellTrackTextButton_InitializeMenuArray()
	local info = { }
	table.insert(info, { text = "   "..BINDING_HEADER_TELLTRACKHEADER, isTitle = 1 })
	table.insert(info, { text = TELLTRACK_WHISPER, func = TellTrack_Menu_Whisper })
	table.insert(info, { text = TELLTRACK_SHOW_CONVERSATION, func = TellTrack_Menu_Show_Conversation })
	table.insert(info, { text = TELLTRACK_WHO, func = TellTrack_Menu_Who })
	table.insert(info, { text = TELLTRACK_GRPINV, func = TellTrack_Menu_GroupeInvite })
	table.insert(info, { text = TELLTRACK_ADDFRIEND, func = TellTrack_Menu_AddToFriend })
	table.insert(info, { text = TELLTRACK_DELETE, func = TellTrack_Menu_Delete })
	table.insert(info, { text = "|cFFCCCCCC--------------|r", disabled = 1, notClickable = 1 })
	table.insert(info, { text = TELLTRACK_DELETE_ALL, func = TellTrack_Menu_DeleteAll })
	table.insert(info, { text = TELLTRACK_INVERT, func = TellTrack_Toggle_Inverted })
	table.insert(info, { text = TELLTRACK_SHOW_ALL_WHISPERS, func = TellTrack_ShowAllWhispers })
	table.insert(info, { text = TELLTRACK_CREATE_CHATFRAME, func = function(name) TellTrack:CreateWhisperChatFrame(name) end })
	table.insert(info, { text = "|cFFCCCCCC--------------|r", disabled = 1, notClickable = 1 })
	table.insert(info, { text = TELLTRACK_CANCEL, func = function () end })
	for index, menuLine in pairs(info) do
		UIDropDownMenu_AddButton(menuLine)
	end
end

function TellTrack:TextButton_GetMenuLevels(menu)
	if not menu then return 0 end
	local menuLevels = 1
	for i = 1, #(menu) do
		if menu[i][1] then
			menuLevels = menuLevels + self:TextButton_GetMenuLevels(menu[i])
			break
		end
	end
	return menuLevels
end

function TellTrack:TextButton_OnClick(frame, button)
	local id = frame:GetID()
	if not id then return end
	if TellTrack_Array[self:GetArrayId(id)] == nil then
		return
	end
	TellTrack_ID = id
	if button == "RightButton" then
		ToggleDropDownMenu(1, nil, TellTrackDropDown, "cursor", 0, 0)
	elseif button == "LeftButton" then
		if (IsShiftKeyDown() and IsAltKeyDown()) then
			TellTrack_Menu_Delete()
		elseif (IsShiftKeyDown()) then
			TellTrack_Menu_Who()
		elseif (IsAltKeyDown()) then
			TellTrack_Menu_GroupeInvite()
		elseif (IsControlKeyDown()) then
			TellTrack_Menu_AddToFriend()
		else
			local name = self:GetName(id)
			if (self._options.WhisperFirst == 1) then
				if (SELECTED_DOCK_FRAME.editBox:IsVisible()) and (SELECTED_DOCK_FRAME.editBox:GetAttribute("chatType") == "WHISPER") and (SELECTED_DOCK_FRAME.editBox:GetAttribute("tellTarget") == name) then
					self:ShowConversationWith(name, false)
				else
					self:InitiateWhisperToTarget(name)
				end
			else
				if (SELECTED_DOCK_FRAME == getglobal(self._chat.frame)) and (TellTrack_CurrentConversation.name == name) then
					self:InitiateWhisperToTarget(name)
				else
					self:ShowConversationWith(name, false)
				end
			end
		end
	end
end

function TellTrack:GetArrayId(id)
	local newID
	if ( self._options.InvertedList == 1 ) then
		newID = TellTrack_ArrayMaxSize - (id + TellTrack_ArrayOffset) + 2
	else
		newID = (id + TellTrack_ArrayOffset) - 1
	end
	return newID
end

function TellTrack:EraseByID(id)
	if id then
		id = self:GetArrayId(id)
		if ( TellTrack_Array ) and ( TellTrack_Array[id] ) then
			if ( ( TellTrackTooltip:IsVisible() ) and ( TellTrack_TooltipSetId == id ) ) then
				TellTrackTooltip:Hide()
			end
			TellTrack_Array[id] = nil
		end
		self:CompressList()
		self:UpdateTellTrackButtonsText()
	end
end

function TellTrack:GetName(id)
	if id then
		id = self:GetArrayId(id)
		if ( ( TellTrack_Array ) and ( TellTrack_Array[id] ) and ( TellTrack_Array[id].name ) ) then
			return TellTrack_Array[id].name
		end
	end
end

function TellTrack:SetTooltip(id)
	local name = self:GetName(id)
	if name then
		TellTrackTooltip:SetText(name)
		TellTrack_TooltipSetId = id
	end
end

-- Yet another function from George Warner, modified a bit to fit my own nefarious purposes. 
-- It can now accept r, g and b specifications, too (leaving out a), as well as handle 255 255 255
function TellTrack:GetColorFormatString(a, r, g, b)
	local percent = false
	if ( ( ( not b ) or ( b <= 1 ) ) and ( a <= 1 ) and ( r <= 1 ) and ( g <= 1) ) then percent = true; end
	if ( ( not b ) and ( a ) and ( r ) and ( g ) ) then b = g; g = r; r = a; if ( percent ) then a = 1; else a = 255; end end
	if ( percent ) then a = a * 255; r = r * 255; g = g * 255; b = b * 255; end
	a = self:GetByteValue(a); r = self:GetByteValue(r); g = self:GetByteValue(g); b = self:GetByteValue(b)
	
	--return format("[c%02X%02X%02X%02X%%s]r", a, r, g, b)
	return format("|c%02X%02X%02X%02X%%s|r", a, r, g, b)
end

function TellTrack:UpdateTellTrackButtonsText()
	local lastSentToNameFormatStr = self:GetColorFormatString(0.2, 1.0, 0.2)
	local lastRecievedFromNameFormatStr = self:GetColorFormatString(1.0, 0.2, 0.2)
	local noNameFormatStr = self:GetColorFormatString(0.4, 0.4, 0.4)
	local id = 0
	for i = 1, TellTrack_ButtonCount do
		local buttonText = getglobal("TellTrack"..i.."Text")
		local formatStr, valueStr
		id = self:GetArrayId(i)
		if ( ( TellTrack_Array[id] ) and ( TellTrack_Array[id].name ) ) then
			valueStr = TellTrack_Array[id].name
			if (TellTrack_OfflinePlayers[valueStr]) then
				formatStr = noNameFormatStr
			elseif( TellTrack_Array[id].sentTo ) then
				formatStr = lastSentToNameFormatStr
			else
				formatStr = lastRecievedFromNameFormatStr
			end
			if (TellTrack_Array[id].unread) and (TellTrack_Array[id].unread > 0) then
				valueStr = "("..TellTrack_Array[id].unread..") "..valueStr
			end
		else
			formatStr = noNameFormatStr
			valueStr = "Empty"
		end
		if ( buttonText ) then
			buttonText:SetText(format(formatStr, valueStr))
			buttonText:Show()
		end
	end
end

--[[ The following function was edited by AnduinLothar (1/15/05)
		to add the following features:

	 The last person whispered is automaticly moved to the bottom
	of the list and everytime you whisper it autoscrolls to the 
	bottom of the list.  Also if the list is full the top name is
	deleted and the new name is appended to the bottom.
        
        1/17/05
	Added recieved/sent boolean for color on button reload.
	
]]--
function TellTrack:HandleMessageSentOrRecieved(target, recieved, msg)
	local tempName = strlower(target)
	local firstEmptySlotIndex = nil
	local previousNameInstance = nil
	for i = 1, TellTrack_ArrayMaxSize do
		if ( ( not firstEmptySlotIndex ) and ( not TellTrack_Array[i] )) then
			firstEmptySlotIndex = i
		elseif( not firstEmptySlotIndex ) then
			if( TellTrack_Array[i].compareName == tempName ) then
				previousNameInstance = i
			end
		end
	end
	if( (not firstEmptySlotIndex) and (not previousNameInstance) ) then
		TellTrack_Array[1] = nil
		self:CompressList()
		firstEmptySlotIndex = TellTrack_ArrayMaxSize
	elseif( (not firstEmptySlotIndex) and previousNameInstance) then
		firstEmptySlotIndex = TellTrack_ArrayMaxSize+1
	end
	for i = 1, TellTrack_ArrayMaxSize do
		if ( previousNameInstance == i and firstEmptySlotIndex == i+1) then
			TellTrack_Array[i].sentTo = not recieved
			if (not TellTrack_Array[i].unread) then
				TellTrack_Array[i].unread = 0
			end
			if recieved and self._chat.frame then
				if (not getglobal(self._chat.frame):IsVisible()) or (TellTrack_CurrentConversation.name ~= target and not TellTrack_CurrentConversation.showAll) then
					TellTrack_Array[i].unread = TellTrack_Array[i].unread + 1
				end
			end
			firstEmptySlotIndex = firstEmptySlotIndex-1
			if (self._options.InvertedList == 1) then
				if( i < TellTrack_ArrayMaxSize-TellTrack_ButtonCount and (TellTrack_ArrayMaxSize-i+1) ~= TellTrack_ArrayOffset ) then
					self:ChangeArrayOffset(TellTrack_ArrayMaxSize-i+1)
				end
			else
				if( i > TellTrack_ButtonCount and (i-TellTrack_ButtonCount+1) ~= TellTrack_ArrayOffset ) then
					self:ChangeArrayOffset(i-TellTrack_ButtonCount+1)
				end
			end
		elseif ( previousNameInstance == i and firstEmptySlotIndex ~= i+1) then
			TellTrack_Array[i] = nil
			self:CompressList()
			firstEmptySlotIndex = firstEmptySlotIndex-1
		elseif ( firstEmptySlotIndex == i ) then
			TellTrack_Array[i] = {}
			TellTrack_Array[i].name = target
			TellTrack_Array[i].compareName = tempName
			TellTrack_Array[i].sentTo = not recieved
			if (not TellTrack_Array[i].unread) then
				TellTrack_Array[i].unread = 0
			end
			if recieved and self._chat.frame then
				if (not getglobal(self._chat.frame):IsVisible()) or (TellTrack_CurrentConversation.name ~= target and not TellTrack_CurrentConversation.showAll) then
					TellTrack_Array[i].unread = TellTrack_Array[i].unread + 1
				end
			end
			if (self._options.InvertedList == 1) then
				if( i < TellTrack_ArrayMaxSize-TellTrack_ButtonCount and (TellTrack_ArrayMaxSize-i+1) ~= TellTrack_ArrayOffset ) then
					self:ChangeArrayOffset(TellTrack_ArrayMaxSize-i+1)
				end
			else
				if( i > TellTrack_ButtonCount and (i-TellTrack_ButtonCount+1) ~= TellTrack_ArrayOffset ) then
					self:ChangeArrayOffset(i-TellTrack_ButtonCount+1)
				end
			end
		end
	end
	self:UpdateTellTrackButtonsText()
	-- saves the names to the list
	self:SaveNames()
end

-- thanks again to Lash for the idea and the pushing of this function :)
function TellTrack:CompressList()
	local index
	local otherArray = {}
	index = 1
	if TellTrack_Array then
		for k, v in pairs(TellTrack_Array) do
			otherArray[index] = v
			index = index + 1
		end
	end
	TellTrack_Array = otherArray
	self:UpdateTellTrackButtonsText()
end


function TellTrack:ChangeArrayOffset(offset)
	local capSize = TellTrack_ArrayMaxSize - TellTrack_ButtonCount+1
	if ( offset <= 0 ) then
		offset = 1
	end
	if ( offset > capSize ) then
		offset = capSize
	end
	TellTrack_ArrayOffset = offset
	self:UpdateTellTrackButtonsText()
end


function TellTrack:OnMouseWheel(frame, v)
	local v = v < 0 and 1 or (v > 0 and -1 or 0)
	self:ChangeArrayOffset(TellTrack_ArrayOffset -  v)
end


-- returns the index name for the load/save funcs
function TellTrack:GetListIndex()
	local firstString = GetCVar("realmName")
	local secondString = UnitRace("player")
	if not firstString or not secondString then
		return nil
	end
	
	local hordeClasses, isHorde = { "Orc", "Tauren", "Troll", "Undead" }
	for k, v in pairs(hordeClasses) do
		if secondString == v then
			isHorde = rtue
		end
	end	
	if isHorde then
		secondString = "Horde"
	else
		secondString = "Alliance"
	end
	return format("%s_%s", firstString, secondString)
end

function TellTrack:LoadNames()
	local index = self:GetListIndex()
	local list = self._db.Log[index]
	if not list then return end
	--1/20/05 now aditionally loads sentTo instead of just names. This allows loading of 'recieved/sent' status.
	for k, v in ipairs(list) do
		if(type(v) == "table") then
			if(v.name and v.sentTo ~= nil) then
				self:HandleMessageSentOrRecieved(v.name, not v.sentTo)
			elseif(v.name) then
				self:HandleMessageSentOrRecieved(v.name, false)
			end
		elseif(type(v) == "string") then
			self:HandleMessageSentOrRecieved(v, false); --Support for previous TellTrack Saved Lists (all are marked as sent)
		end
		if TellTrack_Array[k] then
			TellTrack_Array[k].unread = 0
		end
	end
end

function TellTrack:SaveNames()
	local index = self:GetListIndex()
	list = TellTrack_Array
    self._db.Log[index] = list
end

-- helper function - returns value as a byte
function TellTrack:GetByteValue(pValue)
	local value = tonumber(pValue)
	if ( value <= 0 ) then return 0; end
	if ( value >= 255 ) then return 255; end
	return value
end


function TellTrack:ModifyButtonCount(height)
	height = height - 44;   --insets and arrow buttons
	local newCount = floor((height)/20)
	if ( newCount ~= TellTrack_ButtonCount ) then
		TellTrack_ButtonCount = newCount
		for i=1, TellTrack_ArrayMaxSize do
			local thisButton = getglobal("TellTrack"..i)
			if not thisButton:IsVisible() and i <= TellTrack_ButtonCount then
				thisButton:Show()
			elseif thisButton:IsVisible() and i > TellTrack_ButtonCount then
				thisButton:Hide()
			end
		end
	end
	local offset = floor((height-TellTrack_ButtonCount*20+20)/2)
	TellTrackFrameArrowUpButton:SetHeight(offset)
	TellTrackFrameArrowDownButton:SetHeight(offset)
end

function TellTrack:OnSizeChanged( frame )
	local width = frame:GetWidth()-24
	TellTrackFrameArrowUpButton:SetWidth(width)
	for i=1, TellTrack_ArrayMaxSize do
		getglobal("TellTrack"..i):SetWidth(width)
	end
	TellTrackFrameArrowDownButton:SetWidth(width)
	self:ModifyButtonCount(frame:GetHeight())
	self:ChangeArrayOffset(TellTrack_ArrayOffset)
	self:UpdateTellTrackButtonsText()
end


function TellTrack:Register_Khaos()
	local optionSet = {
		id="TellTrack",
		text=TELLTRACK_CONFIG_HEADER,
		helptext=TELLTRACK_CONFIG_HEADER_INFO,
		difficulty=1,
		options={
			{
				id="Header",
				text=TELLTRACK_CONFIG_HEADER,
				helptext=TELLTRACK_CONFIG_HEADER_INFO,
				type=K_HEADER,
				difficulty=1,
			},
			{
				id="TellTrackEnable",
				type=K_TEXT,
				text=TELLTRACK_ENABLED,
				helptext=TELLTRACK_ENABLED_INFO,
				callback=function(state) if (state.checked) then TellTrack_Toggle_Enabled(1) else TellTrack_Toggle_Enabled(0) end end,
				feedback=function(state) if (state.checked) then return TELLTRACK_CHAT_ENABLED; else return TELLTRACK_CHAT_DISABLED; end end,
				check=true,
				default={checked=true},
				disabled={checked=false},
			},
			{
				id="TellTrackInvert",
				type=K_TEXT,
				text=TELLTRACK_INVERTED,
				helptext=TELLTRACK_INVERTED_INFO,
				callback=function(state) if (state.checked) then TellTrack_Toggle_Inverted(1) else TellTrack_Toggle_Inverted(0) end end,
				feedback=function(state) if (state.checked) then return TELLTRACK_CHAT_INVERTED; else return TELLTRACK_CHAT_NORMALIZED; end end,
				check=true,
				default={checked=false},
				disabled={checked=false},
			},
			{
				id="TellTrackDontSaveList",
				type=K_TEXT,
				text=TELLTRACK_DONTSAVELIST,
				helptext=TELLTRACK_DONTSAVELIST_INFO,
				callback=function(state) if (state.checked) then TellTrack_Toggle_DontSaveList(1) else TellTrack_Toggle_DontSaveList(0) end end,
				feedback=function(state) if (state.checked) then return TELLTRACK_CHAT_DONTSAVELIST; else return TELLTRACK_CHAT_SAVELIST; end end,
				check=true,
				default={checked=false},
				disabled={checked=false},
			},
			{
				id="TellTrackAutoCreate",
				type=K_TEXT,
				text=TELLTRACK_AUTOCREATE,
				helptext=TELLTRACK_AUTOCREATE_INFO,
				callback=function(state) if (state.checked) then TellTrack_Toggle_AutoCreate(1) else TellTrack_Toggle_AutoCreate(0) end end,
				feedback=function(state) if (state.checked) then return TELLTRACK_CHAT_AUTOCREATE; else return TELLTRACK_CHAT_MANUALWHISPERFRAME; end end,
				check=true,
				default={checked=true},
				disabled={checked=false},
			},
			{
				id="TellTrackHideOtherWhispers",
				type=K_TEXT,
				text=TELLTRACK_HIDEOTHERWHISPERS,
				helptext=TELLTRACK_HIDEOTHERWHISPERS_INFO,
				callback=function(state) if (state.checked) then TellTrack_Toggle_HideOtherWhispers(1) else TellTrack_Toggle_HideOtherWhispers(0) end end,
				feedback=function(state) if (state.checked) then return TELLTRACK_CHAT_HIDEOTHERWHISPERS; else return TELLTRACK_CHAT_SHOWOTHERWHISPERS; end end,
				check=true,
				default={checked=false},
				disabled={checked=false},
			},
			{
				id="TellTrackTimeStamps",
				type=K_TEXT,
				text=TELLTRACK_TIMESTAMPS,
				helptext=TELLTRACK_TIMESTAMPS_INFO,
				callback=function(state) if (state.checked) then TellTrack_Toggle_TimeStamps(1) else TellTrack_Toggle_TimeStamps(0) end end,
				feedback=function(state) if (state.checked) then return TELLTRACK_CHAT_TIMESTAMPS; else return TELLTRACK_CHAT_NOTIMESTAMPS; end end,
				check=true,
				default={checked=false},
				disabled={checked=false},
			},
			{
				id="TellTrackWhisperFirst",
				type=K_TEXT,
				text=TELLTRACK_WHISPERFIRST,
				helptext=TELLTRACK_WHISPERFIRST_INFO,
				callback=function(state) if (state.checked) then TellTrack_Toggle_WhisperFirst(1) else TellTrack_Toggle_WhisperFirst(0) end end,
				feedback=function(state) if (state.checked) then return TELLTRACK_CHAT_WHISPERFIRST; else return TELLTRACK_CHAT_LOGFIRST; end end,
				check=true,
				default={checked=false},
				disabled={checked=false},
			},
			{
				id="TellTrackHideBorder",
				type=K_TEXT,
				text=TELLTRACK_HIDEBORDER,
				helptext=TELLTRACK_HIDEBORDER_INFO,
				callback=function(state) if (state.checked) then TellTrack_Toggle_HideBorder(1) else TellTrack_Toggle_HideBorder(0) end end,
				feedback=function(state) if (state.checked) then return TELLTRACK_CHAT_HIDEBORDER; else return TELLTRACK_CHAT_SHOWBORDER; end end,
				check=true,
				default={checked=false},
				disabled={checked=false},
			},
		},
	}
	Khaos.registerOptionSet(
		"chat",
		optionSet
	)
	TellTrack_Khaos_Registered = 1
end

function TellTrack:Register_Satellite()
	Satellite.registerSlashCommand(
		{
			id="TellTrack",
			commands = {"/telltrack","/ttrack",},
			onExecute = SlashCmdList["TELLTRACK"],
			helpText = TELLTRACK_CHAT_COMMAND_INFO,
		}
	)
	Satellite.registerSlashCommand(
		{
			id="Retell",
			commands = {"/re", "/retell",},
			onExecute = TellTrack_WhisperToPreviousTarget,
			helpText = BINDING_NAME_TELLTRACK_RETELL,
		}
	)
	TellTrack_Satellite_Registered = 1
end
